import numpy as np
from pysimenv.core.system import MultipleSystem
from pysimenv.multicopter.model import MulticopterDynamic
from pysimenv.multicopter.control import FLVelControl
from pysimenv.core.simulator import Simulator


class FLVelTracking(MultipleSystem):
    def __init__(self):
        super(FLVelTracking, self).__init__()

        m = 0.5
        J = np.diag([4.85, 4.85, 8.81])*1e-3

        pos = np.zeros(3)
        vel = np.zeros(3)
        R_iv = np.identity(3)
        omega = np.zeros(3)
        self.quadrotor = MulticopterDynamic([pos, vel, R_iv, omega], m, J)

        k_p_att = np.array([1600., 1600., 1600])
        k_d_att = np.array([80., 80., 80])
        k_p_vel = np.array([5., 5., 5.])
        self.vel_control = FLVelControl(m, J, k_p_att, k_d_att, k_p_vel)

        self.attach_sim_objects([self.quadrotor, self.vel_control])

    # implement
    def _forward(self, v_d: np.ndarray = np.zeros(3)):
        v = self.quadrotor.vel
        eta = self.quadrotor.euler_ang
        omega = self.quadrotor.ang_vel

        u = self.vel_control.forward(v, eta, omega, v_d)
        self.quadrotor.forward(u=u)


def main():
    print("== Test for feedback linearization velocity controller == ")
    v_d = np.array([0.5, 0., 1.])
    model = FLVelTracking()
    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, v_d=v_d)
    model.quadrotor.default_plot(show=True)


if __name__ == "__main__":
    main()




import numpy as np
from pysimenv.core.system import MultipleSystem
from pysimenv.multicopter.model import MulticopterDynamic
from pysimenv.multicopter.control import QuaternionAttControl
from pysimenv.common.orientation import euler_angles_to_rotation
from pysimenv.common.model import FlatEarthEnv
from pysimenv.core.simulator import Simulator


class Model(MultipleSystem):
    def __init__(self):
        super(Model, self).__init__()

        m = 1.3
        J = np.diag([0.0119, 0.0119, 0.0219])

        pos = np.array([1., 2., 0.])
        vel = np.zeros(3)
        R_vi = euler_angles_to_rotation(np.deg2rad([10., 10., 20.]))
        R_iv = np.transpose(R_vi)
        omega = np.array([0., 0., 1.])
        self.quadrotor = MulticopterDynamic([pos, vel, R_iv, omega], m, J)

        Q = np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2
        R = 1e-4*np.identity(3)
        K = QuaternionAttControl.gain(Q, R)
        self.att_control = QuaternionAttControl(J, K)

        self.attach_sim_objects([self.quadrotor, self.att_control])

    def _forward(self, q_d: np.ndarray, omega_d: np.ndarray = np.zeros(3)):
        q = self.quadrotor.quaternion
        omega = self.quadrotor.ang_vel

        f = self.quadrotor.m*FlatEarthEnv.grav_accel
        tau = self.att_control.forward(q, omega, q_d, omega_d)
        self.quadrotor.forward(u=np.array([f, tau[0], tau[1], tau[2]]))


def main():
    print("== Test for attitude control based on quaternion ==")
    q_d = np.array([1., 0., 0., 0.])
    model = Model()
    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, q_d=q_d)
    model.quadrotor.default_plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from pysimenv.core.system import MultipleSystem
from pysimenv.missile.model import PlanarMissile2dof, PlanarManVehicle2dof
from pysimenv.missile.guidance import PurePNG2dim, IACBPNG
from pysimenv.missile.util import RelKin2dim, CloseDistCond, miss_distance, closest_instant, lin_interp


class Engagement2dim(MultipleSystem):
    def __init__(self, missile: PlanarMissile2dof, target: PlanarManVehicle2dof):
        super(Engagement2dim, self).__init__()
        self.missile = missile
        self.target = target
        self.rel_kin = RelKin2dim()
        self.close_dist_cond = CloseDistCond(r_threshold=10.0)

        self.attach_sim_objects([self.missile, self.target])

    # override
    def reset(self):
        super(Engagement2dim, self).reset()
        self.close_dist_cond.reset()

    # implement
    def initialize(self):
        x_M = self.missile.state['x']
        x_T = self.target.state['x']
        self.rel_kin.evaluate(x_M, x_T)

    # implement
    def forward(self):
        x_M = self.missile.state['x']
        x_T = self.target.state['x']
        self.rel_kin.evaluate(x_M, x_T)
        self.close_dist_cond.evaluate(self.rel_kin.r)
        self.target.forward()

    # implement
    def check_stop_condition(self) -> Tuple[bool, int]:
        to_stop = False

        missile_stop, _ = self.missile.check_stop_condition()
        if self.intercepted():  # probable interception
            to_stop = True
            self.flag = 1

        if missile_stop:  # stop due to the missile
            to_stop = True
            self.flag = 2

        if self.is_out_of_view():  # out of field-of-view limit
            to_stop = True
            self.flag = 3

        return to_stop, self.flag

    def intercepted(self) -> bool:
        return self.close_dist_cond.check()

    def is_out_of_view(self) -> bool:
        sigma = self.rel_kin.sigma
        return self.missile.is_out_of_view(sigma)

    def miss_distance(self) -> float:
        x_M = self.missile.history('x')
        x_T = self.target.history('x')

        p_M = x_M[:, 0:2]
        p_T = x_T[:, 0:2]
        d_miss = miss_distance(p_M, p_T)
        return d_miss

    def state_on_closest_instant(self):
        x_M = self.missile.history('x')
        x_T = self.target.history('x')

        p_M = x_M[:, 0:2]
        p_T = x_T[:, 0:2]
        index_close, xi_close = closest_instant(p_M, p_T)
        x_M_close = lin_interp(x_M[index_close], x_M[index_close + 1], xi_close)
        x_T_close = lin_interp(x_T[index_close], x_T[index_close + 1], xi_close)
        return x_M_close, x_T_close

    def report(self):
        self.missile.report()
        if self.flag == 1:
            print("[engagement] The target has been intercepted!")
        else:
            print("[engagement] The target has been missed!")

        d_miss = self.miss_distance()
        x_M_close, x_T_close = self.state_on_closest_instant()

        print("[engagement] Miss distance: {:.6f} (m)".format(d_miss))
        print("[engagement] Missile state on the closest instant: {:.2f}(m), {:.2f}(m), {:.2f}(m/s), {:.2f}(deg) \n".
              format(x_M_close[0], x_M_close[1], x_M_close[2], np.rad2deg(x_M_close[3]))
              )
        print("[engagement] Target state on the closest instant: {:.2f}(m), {:.2f}(m), {:.2f}(m/s), {:.2f}(deg) \n".
              format(x_T_close[0], x_T_close[1], x_T_close[2], np.rad2deg(x_T_close[3]))
              )

    def plot(self):
        fig_axs = dict()

        fig_ax = self.missile.plot_path()
        self.target.plot_path(fig_ax)

        fig1, ax1 = fig_ax['fig'], fig_ax['ax']
        fig1.suptitle("2-dim flight path")
        fig_axs['path'] = {'fig': fig1, 'ax': ax1}

        # plots for rel. kin.
        time_list = self.missile.history('t')
        x_M_list = self.missile.history('x')
        x_T_list = self.target.history('x')

        relKin = RelKin2dim()
        r_list = []
        sigma_list = []
        lam_list = []
        omega_list = []

        zem_list = []

        for i in range(x_M_list.shape[0]):
            x_M = x_M_list[i, :]
            x_T = x_T_list[i, :]

            relKin.evaluate(x_M, x_T)

            r_list.append(relKin.r)
            sigma_list.append(relKin.sigma)
            lam_list.append(relKin.lam)
            omega_list.append(relKin.omega)

            zem_list.append(relKin.zem)

        r_list = np.array(r_list)
        sigma_list = np.array(sigma_list)
        lam_list = np.array(lam_list)
        omega_list = np.array(omega_list)

        zem_list = np.array(zem_list)

        fig2, ax2 = plt.subplots(2, 2)
        fig2.tight_layout()
        ax2_1 = ax2[0, 0]
        ax2_1.set_title("Rel. dist")
        ax2_1.plot(time_list[:-1], r_list[:-1], label="Rel. dist")
        ax2_1.set_xlabel("Time (s)")
        ax2_1.set_ylabel("Rel. dist (m)")
        ax2_1.grid()
        ax2_1.legend()

        ax2_2 = ax2[0, 1]
        ax2_2.set_title("Look angle")
        ax2_2.plot(time_list[:-1], np.rad2deg(sigma_list[:-1]), label="look angle")
        ax2_2.set_xlabel("Time (s)")
        ax2_2.set_ylabel("Look angle (deg)")
        ax2_2.grid()
        ax2_2.legend()

        ax2_3 = ax2[1, 0]
        ax2_3.set_title("LOS angle")
        ax2_3.plot(time_list[:-1], np.rad2deg(lam_list[:-1]), label="LOS angle")
        ax2_3.set_xlabel("Time (s)")
        ax2_3.set_ylabel("LOS angle (deg)")
        ax2_3.grid()
        ax2_3.legend()

        ax2_4 = ax2[1, 1]
        ax2_4.set_title("LOS rate")
        ax2_4.plot(time_list[:-1], np.rad2deg(omega_list[:-1]), label="LOS rate")
        ax2_4.set_xlabel("Time (s)")
        ax2_4.set_ylabel("LOS rate (deg/s)")
        ax2_4.grid()
        ax2_4.legend()

        fig_axs['Rel. Kin.'] = {'fig': fig2, 'ax': ax2}

        fig3, ax3 = plt.subplots()
        ax3.set_title("ZEM")
        ax3.plot(time_list[:-1], zem_list[:-1], label="ZEM")
        ax3.set_xlabel("Time (s)")
        ax3.set_ylabel("ZEM (m)")
        ax3.grid()
        ax3.legend()

        fig_axs['Rel. Kin. add.'] = {'fig': fig3, 'ax': ax3}

        return fig_axs


class PurePNG2dimEngagement(Engagement2dim):
    def __init__(self, missile: PlanarMissile2dof, target: PlanarManVehicle2dof):
        super(PurePNG2dimEngagement, self).__init__(missile, target)

        self.pure_png = PurePNG2dim(N=3.0)
        self.attach_sim_objects([self.pure_png])

    # implement
    def forward(self):
        super(PurePNG2dimEngagement, self).forward()
        V_M = self.missile.V
        omega = self.rel_kin.omega

        a_M = self.pure_png.forward(V_M, omega)
        self.missile.forward(a_M=np.array([0, a_M]))


class IACBPNGEngagement(Engagement2dim):
    def __init__(self, missile: PlanarMissile2dof, target: PlanarManVehicle2dof, bpng: IACBPNG):
        super(IACBPNGEngagement, self).__init__(missile, target)
        self.bpng = bpng
        self.attach_sim_objects([self.bpng])

    def forward(self):
        super(IACBPNGEngagement, self).forward()
        a_y_cmd = self.bpng.forward(self.missile, self.target, self.rel_kin)
        self.missile.forward(a_M=np.array([0., a_y_cmd]))




import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile2dof, PlanarNonManVehicle2dof
from pysimenv.missile.engagement import PurePNG2dimEngagement


print("== Two dimensional pure PNG engagement for a stationary target ==")

dt = 0.01
final_time = 30.
missile = PlanarMissile2dof(
    [-5000., 3000., 300., np.deg2rad(-15.)]
)
target = PlanarNonManVehicle2dof(
    [0., 0., 20., 0.]
)
missile.name = "missile"
target.name = "target"

# Method 1
model = PurePNG2dimEngagement(missile, target)
simulator = Simulator(model)
simulator.propagate(dt, final_time, True)

model.plot()
model.report()
model.save_log_file('./data/planar_png/')
# model.load_log_file('./data/planar_png/')
model.plot()

plt.show()



import numpy as np
import scipy
from pyquaternion import Quaternion
from typing import Union, Tuple
from pysimenv.core.base import StaticObject
from pysimenv.common.model import FlatEarthEnv
from pysimenv.common.orientation import quaternion_to_axis_angle


class FLVelControl(StaticObject):
    """
    Feedback linearization velocity control
    Reference: H. Voos, "Nonlinear Control of a Quadrotor Micro-UAV using Feedback-Linearization",
    Proceedings of the 2009 IEEE International Conference on Mechatronics, 2009.
    """
    def __init__(self, m: float, J: np.ndarray,
                 k_p_att: np.ndarray, k_d_att: np.ndarray, k_p_vel: np.ndarray,
                 interval: Union[int, float] = -1):
        super(FLVelControl, self).__init__(interval=interval)
        self.m = m
        self.J = J  # (3, 3) array
        self.k_p_att = k_p_att  # (3,) array
        self.k_d_att = k_d_att  # (3,) array
        self.k_p_vel = k_p_vel  # (3,) array

    # implement
    def _forward(self, v: np.ndarray, eta: np.ndarray, omega: np.ndarray, v_d: np.ndarray = np.zeros(3)):
        """
        :param v: velocity
        :param eta:
        :param omega:
        :param v_d:
        :return:
        """
        u_tilde = self.k_p_vel*(v_d - v)
        u_t_1, u_t_2, u_t_3 = u_tilde[:]
        g = FlatEarthEnv.grav_accel

        f = self.m*np.sqrt(u_t_1**2 + u_t_2**2 + (u_t_3 - g)**2)
        phi_d = np.arcsin(self.m/f*u_t_2)
        theta_d = np.arctan(u_t_1/(u_t_3 - g))
        psi_d = 0.

        eta_d = np.array([phi_d, theta_d, psi_d])
        u_star = self.k_p_att*(eta_d - eta) - self.k_d_att*omega

        J_x, J_y, J_z = np.diag(self.J)[:]
        J_1 = (J_y - J_z)/J_x
        J_2 = (J_z - J_x)/J_y
        J_3 = (J_x - J_y)/J_z

        p, q, r = omega[:]
        tau_x = J_x*(-q*r*J_1 + u_star[0])
        tau_y = J_y*(-p*r*J_2 + u_star[1])
        tau_z = J_z*(-p*q*J_3 + u_star[2])

        u = np.array([f, tau_x, tau_y, tau_z])
        return u


class QuaternionAttControl(StaticObject):
    """
    Attitude control based on quaternion
    Reference: J. Carino, H. Abaunza and P. Castillo,
    "Quadrotor Quaternion Control",
    2015 International Conference on Unmanned Aircraft Systems (ICUAS), 2015.
    """
    def __init__(self, J: np.ndarray, K: np.ndarray, interval: Union[int, float] = -1):
        super(QuaternionAttControl, self).__init__(interval=interval)
        self.J = J  # (3, 3) array
        self.K = K  # (3, 6) array

    # implement
    def _forward(self, q: np.ndarray, omega: np.ndarray, q_d: np.ndarray, omega_d: np.ndarray = np.zeros(3))\
            -> np.ndarray:
        a, phi = quaternion_to_axis_angle(q)
        theta = a*phi
        x = np.hstack((theta, omega))

        a_d, phi_d = quaternion_to_axis_angle(q_d)
        theta_d = a_d*phi_d
        x_d = np.hstack((theta_d, omega_d))

        u = -self.K.dot(x - x_d)
        tau = self.J.dot(u) + np.cross(omega, self.J.dot(omega))
        return tau

    @classmethod
    def gain(cls, Q: np.ndarray, R: np.ndarray) -> np.ndarray:
        A = np.zeros((6, 6))
        A[0:3, 3:6] = np.identity(3)

        B = np.zeros((6, 3))
        B[3:6, :] = np.identity(3)

        P = scipy.linalg.solve_continuous_are(A, B, Q, R)
        K = np.linalg.inv(R).dot(B.transpose().dot(P))
        return K


class QuaternionPosControl(StaticObject):
    e3 = np.array([0., 0., 1.])
    n = np.array([0., 0., -1.])

    def __init__(self, m: float, K: np.ndarray, interval: Union[int, float] = -1):
        super(QuaternionPosControl, self).__init__(interval=interval)
        self.m = m
        self.K = K

    # implement
    def _forward(self, p: np.ndarray, v: np.ndarray, p_d: np.ndarray, v_d: np.ndarray = np.zeros(3))\
            -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x = np.hstack((p, v))
        x_d = np.hstack((p_d, v_d))
        u = -self.K.dot(x - x_d)

        u_pd = u - FlatEarthEnv.grav_accel*self.e3
        f = self.m*np.linalg.norm(u_pd)

        r_d = np.hstack((
            np.array([self.n.dot(u_pd) + np.linalg.norm(u_pd)]),
            np.cross(self.n, u_pd)
        ))

        q_d = r_d/np.linalg.norm(r_d)

        x_dot = np.hstack((v, u))
        u_pd_dot = -self.K.dot(x_dot)
        r_d_dot = np.hstack((
            np.array([self.n.dot(u_pd_dot) + u_pd.dot(u_pd_dot)/np.linalg.norm(u_pd)]),
            np.cross(self.n, u_pd_dot)
        ))
        q_d_dot = r_d_dot/np.linalg.norm(r_d) + r_d*(
            -r_d.dot(r_d_dot)/np.linalg.norm(r_d)**3)

        omega_d = 2*Quaternion(q_d).conjugate*Quaternion(q_d_dot)
        omega_d = omega_d.elements[1:4]
        return f, q_d, omega_d

    @classmethod
    def gain(cls, Q: np.ndarray, R: np.ndarray):
        A = np.zeros((6, 6))
        A[0:3, 3:6] = np.identity(3)

        B = np.zeros((6, 3))
        B[3:6, :] = np.identity(3)

        P = scipy.linalg.solve_continuous_are(A, B, Q, R)
        K = np.linalg.inv(R).dot(B.transpose().dot(P))
        return K



import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, Tuple
from pysimenv.core.base import ArrayType
from pysimenv.core.system import DynSystem
from pysimenv.common.util import wrap_to_pi


class PlanarManVehicle2dof(DynSystem):
    """
    A 2-dof model for a maneuvering vehicle in the plane
    state x: [p_x, p_y, V, gamma]
    control u: [a_x, a_y]
    where [p_x, p_y] is the position expressed in the inertial frame and
    [a_x, a_y] is the acceleration expressed in the velocity frame
    """
    def __init__(self, initial_state: ArrayType):
        super(PlanarManVehicle2dof, self).__init__(initial_states={'x': initial_state})
        self.name = "planar_man_vehicle"

        def angle_correction_fun(state: np.ndarray) -> np.ndarray:
            state[3] = wrap_to_pi(state[3])  # gamma should be in [-pi, pi] rad
            return state
        self.state_vars['x'].attach_correction_fun(angle_correction_fun)

    # implement
    def _deriv(self, x: np.ndarray, u: Optional[ArrayType] = None):
        """
        :param x: [p_x, p_y, V, gamma]
        :param u: [a_x, a_y]
        :return: x_dot
        """
        if u is None:
            u = np.array([0., 0.])
        V = x[2]
        gamma = x[3]

        a_x = u[0]
        a_y = u[1]

        p_x_dot = V*np.cos(gamma)
        p_y_dot = V*np.sin(gamma)
        V_dot = a_x

        if abs(V) < 1e-4:
            gamma_dot = 0
        else:
            gamma_dot = a_y/V

        return {'x': np.array([p_x_dot, p_y_dot, V_dot, gamma_dot])}

    @property
    def p(self) -> np.ndarray:
        """
        :return: [p_x, p_y]
        """
        return self.state['x'][0:2]

    @property
    def v(self) -> np.ndarray:
        """
        :return: [v_x, v_y]
        """
        x = self.state['x']
        V = x[2]
        gamma = x[3]

        return np.array([[V*np.cos(gamma), V*np.sin(gamma)]])

    @property
    def V(self) -> float:
        """
        :return: V
        """
        return self.state['x'][2]

    @property
    def gamma(self) -> float:
        """
        :return: gamma
        """
        return self.state['x'][3]

    def plot(self):
        var_keys = {'x', 'u'}
        var_ind_dict = {'x': [0, 1, 2, 3], 'u': [0, 1]}
        var_names_dict = {
            'x': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)'],
            'u': ['a_x (m/s**2)', 'a_y (m/s**2)']
        }
        fig_axs = self.default_plot(var_keys=var_keys, var_ind_dict=var_ind_dict, var_names_dict=var_names_dict)
        return fig_axs

    def plot_path(self, fig_ax=None):
        if fig_ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("p_x (m)")
            ax.set_ylabel("p_y (m)")
            ax.grid()
        else:
            fig = fig_ax['fig']
            ax = fig_ax['ax']

        state_list = self.history('x')
        pos_list = state_list[:, 0:2]
        if abs(self.V) < 1e-4:
            ax.plot(pos_list[0, :], pos_list[0, :], marker='o', label=self.name)
        else:
            ax.plot(pos_list[:, 0], pos_list[:, 1], label=self.name + " path")
        ax.set_aspect('equal')
        ax.legend()

        plt.draw()
        plt.pause(0.01)
        return {'fig': fig, 'ax': ax}


class PlanarNonManVehicle2dof(PlanarManVehicle2dof):
    def __init__(self, initial_state: ArrayType):
        super(PlanarNonManVehicle2dof, self).__init__(initial_state)
        self.name = "planar_non_man_vehicle"

    # override
    def _deriv(self, x: np.ndarray, **kwargs):
        return super(PlanarNonManVehicle2dof, self)._deriv(x, np.zeros(2))

    # override
    def plot(self):
        var_keys = {'x'}
        var_ind_dict = {'x': [0, 1, 2, 3]}
        var_names_dict = {
            'x': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)']
        }
        fig_axs = self.default_plot(var_keys, var_ind_dict, var_names_dict)
        return fig_axs


class PlanarMissile2dof(PlanarManVehicle2dof):
    def __init__(self, initial_state: ArrayType):
        super(PlanarMissile2dof, self).__init__(initial_state)
        self.name = "missile"
        self.fov_limit = np.inf
        self.acc_limit = np.array([
            [-np.inf, -np.inf],
            [np.inf, np.inf]
        ])
        self.ground_elevation = -np.inf

    # override
    def _forward(self, a_M):
        a_M = np.clip(a_M, self.acc_limit[0], self.acc_limit[1])
        super(PlanarMissile2dof, self)._forward(u=a_M)

    # implement
    def check_stop_condition(self) -> Tuple[bool, int]:
        to_stop = False
        if self.is_stalled():
            to_stop = True
            self.flag = 1
        if self.is_collided():
            to_stop = True
            self.flag = 2
        return to_stop, self.flag

    def is_stalled(self) -> bool:
        return self.V < 10  # when the speed is less than 10m/s

    def is_collided(self) -> bool:
        return self.p[1] < self.ground_elevation - 0.5

    def is_out_of_view(self, sigma: float) -> bool:
        return abs(sigma) > self.fov_limit

    def report(self):
        np.set_printoptions(precision=2, suppress=True)
        print("[{:s}] Final state [m, m, m/s, deg]:".format(self.name))

        final_state = self.state['x'].copy()
        final_state[3] = np.rad2deg(final_state[3])
        print(final_state)
        if self.flag == 0:
            print("[{:s}] Status: normal".format(self.name))
        elif self.flag == 1:
            print("[{:s}] Status: stalled".format(self.name))
        elif self.flag == 2:
            print("[{:s}] Status: collided".format(self.name))



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile2dof, PlanarNonManVehicle2dof
from pysimenv.missile.guidance import IACBPNG
from pysimenv.missile.engagement import IACBPNGEngagement


dt = 0.01
final_time = 30.

theta_M_0 = np.deg2rad(30.)
theta_T = 0.

missile = PlanarMissile2dof(
    [0., 0., 200., theta_M_0]
)
target = PlanarNonManVehicle2dof(
    [3000., 0., 30., theta_T]
)
missile.name = "missile"
target.name = "target"


theta_M_f = np.deg2rad(-40.)
sigma_max = np.deg2rad(45.)
bpng = IACBPNG(N=3., tau=1., theta_M_0=theta_M_0, theta_M_f=theta_M_f, theta_T=theta_T,
               lam_0=0., sigma_max=sigma_max, b_max=0.35)
model = IACBPNGEngagement(missile, target, bpng)
simulator = Simulator(model)
simulator.propagate(dt, final_time, True)

model.plot()
model.report()
bpng.plot_bias()
bpng.plot_bias_integral()

plt.show()



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarManVehicle2dof, PlanarNonManVehicle2dof


print("== Test for PlanarManVehicle2dof ==")
manVehicle = PlanarManVehicle2dof(
    initial_state=[-5000., 0., 200., np.deg2rad(30.)])

u = np.array([0., -5.])
dt = 0.01
time = 30.

simulator = Simulator(manVehicle)
simulator.propagate(dt, time, True, u=u)

manVehicle.default_plot()
manVehicle.plot_path()
plt.show()

print("== Test for PlanarNonManVehicle2dof ==")
nonManVehicle = PlanarNonManVehicle2dof(
    initial_state=[0., 0., 10., np.deg2rad(175.)]
)
simulator = Simulator(nonManVehicle)
simulator.propagate(dt, time, True)

nonManVehicle.default_plot()
nonManVehicle.plot_path()
plt.show()



import numpy as np
import matplotlib.pyplot as plt
from typing import Union
from pysimenv.core.base import StaticObject
from pysimenv.core.system import MultipleSystem
from pysimenv.common.model import Integrator
from pysimenv.missile.model import PlanarManVehicle2dof
from pysimenv.missile.util import RelKin2dim


class PurePNG2dim(StaticObject):
    def __init__(self, N: float = 3.0, interval: Union[int, float] = -1):
        super(PurePNG2dim, self).__init__(interval=interval)
        self.N = N

    # implement
    def _forward(self, V_M, omega):
        """
        :param V_M: speed of the missile
        :param omega: LOS rate
        :return: acceleration command a_M
        """
        a_M = self.N*V_M*omega
        return a_M


class IACBPNG(MultipleSystem):
    """
    Biased PNG with terminal-angle constraint (impact angle control)
    """
    def __init__(self, N: float, tau: float, theta_M_0: float, theta_M_f: float, theta_T: float, lam_0: float,
                 sigma_max: float = float('inf'), b_max: float = float('inf')):
        super(IACBPNG, self).__init__()

        # Guidance parameters
        self.N = N
        self.tau = tau
        self.theta_M_0 = theta_M_0
        self.theta_M_f = theta_M_f
        self.theta_T = theta_T
        self.lam_0 = lam_0
        self.sigma_max = sigma_max
        self.b_max = b_max

        self.phase = 0  # 0: BPNG, 1: DPP, 2: BPNG
        self.bias_integrator = Integrator(np.array([0.]))
        self.attach_sim_objects([self.bias_integrator])

    @classmethod
    def reference_value(cls, theta_M_0, theta_M_f, theta_T, lam_0, N, V_M, V_T):
        theta_M_0_bar = theta_M_0 - lam_0
        theta_M_f_bar = theta_M_f - lam_0
        theta_T_bar = theta_T - lam_0

        lam_f_bar = np.arctan(
            (V_M*np.sin(theta_M_f_bar) - V_T*np.sin(theta_T_bar))/(V_M*np.cos(theta_M_f_bar) - V_T*np.cos(theta_T_bar))
        )
        B_ref = theta_M_f_bar - theta_M_0_bar - N*lam_f_bar
        return B_ref

    def _forward(self, missile: PlanarManVehicle2dof, target: PlanarManVehicle2dof, rel_kin: RelKin2dim) -> float:
        V_M = missile.V
        V_T = target.V
        omega = rel_kin.omega
        sigma = rel_kin.sigma

        B_ref = IACBPNG.reference_value(
            self.theta_M_0, self.theta_M_f, self.theta_T, self.lam_0, self.N, V_M, V_T)
        B = self.bias_integrator.state['x'][0]

        b_1 = 1./self.tau*(B_ref - B)
        b_2 = (1. - self.N)*omega
        if self.phase == 0:
            # BPNG
            b = b_1
            if np.abs(sigma) > self.sigma_max - np.deg2rad(0.01):
                self.phase = 1
        elif self.phase == 1:
            # DPP
            b = b_2
            if np.abs(b_1) < np.abs(b_2):
                self.phase = 2
        else:
            # BPNG
            b = b_1

        b = np.clip(b, -self.b_max, self.b_max)
        a_png = self.N*V_M*omega
        a_b = V_M*b
        a_M = a_png + a_b

        self.bias_integrator.forward(u=b)
        self._logger.append(t=self.time, B_ref=B_ref, B=B, b=b, a_png=a_png, a_b=a_b)
        return a_M

    def plot_bias(self):
        t = self.history('t')
        b = self.history('b')

        fig, ax = plt.subplots()
        ax.plot(t, b)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Bias, b (rad/s)")
        ax.grid()

    def plot_bias_integral(self):
        t = self.history('t')
        B_ref = self.history('B_ref')
        B = self.history('B')

        fig, ax = plt.subplots()
        ax.plot(t, B_ref, linestyle='--')
        ax.plot(t, B)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Integral Value of Bias, B (rad)")
        ax.grid()



import h5py
import os
import numpy as np
import matplotlib.pyplot as plt
from abc import ABC
from typing import Union, Tuple, Dict, Optional
from pysimenv.core.util import SimClock, Timer
from pysimenv.core.base import SimObject, StateVariable, BaseFunction, ArrayType


class DynObject(SimObject):
    def __init__(self, initial_states: Optional[Dict[str, ArrayType]] = None, interval: Union[int, float] = -1):
        """
        :param kwargs: initial states
        """
        super(DynObject, self).__init__(interval=interval)
        self.name = "base_system"
        self.state_vars: Dict[str, StateVariable] = dict()

        state_dim = 0
        if initial_states is not None:
            for name, initial_state in initial_states.items():
                var = StateVariable(initial_state)
                self.state_vars[name] = var
                state_dim += var.size

        self.num_state_var = len(self.state_vars)
        self.state_dim = state_dim

    def set_state(self, **kwargs) -> None:
        """
        :param kwargs: states
        :return: None
        """
        if len(kwargs) > 0:
            for name, state in kwargs.items():
                self.state_vars[name].apply_state(state)

    def step(self, dt: float, **kwargs) -> None:
        t_0 = self._sim_clock.time

        self._log_timer.forward()
        self.forward(**kwargs)
        for var in self.state_vars.values():
            var.rk4_update_1(dt)

        self._sim_clock.apply_time(t_0 + dt / 2)
        self._log_timer.forward()
        self.forward(**kwargs)
        for var in self.state_vars.values():
            var.rk4_update_2(dt)

        self.forward(**kwargs)
        for var in self.state_vars.values():
            var.rk4_update_3(dt)

        self._sim_clock.apply_time(t_0 + dt - 10 * self._sim_clock.time_res)
        self._log_timer.forward()
        self.forward(**kwargs)
        for var in self.state_vars.values():
            var.rk4_update_4(dt)

        self._sim_clock.apply_time(t_0 + dt)

    def propagate(self, dt: float, time: float, **kwargs):
        assert self._sim_clock is not None, "Attach a sim_clock first!"
        assert self._log_timer is not None, "Attach a log_timer first!"

        iter_num = min(round(time/dt), np.iinfo(np.int32).max)
        for i in range(iter_num):
            to_stop, _ = self.check_stop_condition()
            if to_stop:
                break
            self.step(dt, **kwargs)

        self._log_timer.forward()
        self.forward(**kwargs)

    def default_plot(self, show=False, var_keys=None, var_ind_dict=None, var_names_dict=None):
        if var_keys is None:
            var_keys = list(self._logger.keys())
            var_keys.remove('t')
        if var_ind_dict is None:
            var_ind_dict = dict()
        if var_names_dict is None:
            var_names_dict = dict()

        fig_axs = dict()
        time_list = self.history('t')

        for var_key in var_keys:
            var_list = self.history(var_key)
            if var_list.ndim == 1:
                var_list = np.reshape(var_list, var_list.shape + (1, ))

            if var_key in var_ind_dict:
                ind = var_ind_dict[var_key]
            else:
                ind = list(range(var_list.shape[1]))

            if var_key in var_names_dict:
                names = var_names_dict[var_key]
            else:
                names = [var_key + "_" + str(k) for k in ind]

            subplot_num = len(ind)
            fig, ax = plt.subplots(subplot_num, 1)
            fig_axs[var_key] = {
                'fig': fig,
                'ax': ax
            }
            for i in range(subplot_num):
                if subplot_num == 1:
                    ax_ = ax
                else:
                    ax_ = ax[i]
                ax_.plot(time_list, var_list[:, ind[i]], label="Actual")
                ax_.set_xlabel("Time (s)")
                ax_.set_ylabel(names[i])
                ax_.grid()
                ax_.legend()
            fig.suptitle(var_key)
            fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return fig_axs

    # to be implemented
    def report(self):
        pass

    def save(self, h5file=None, data_group=''):
        data_group = data_group + '/' + self.name
        self._logger.save(h5file, data_group)

    def load(self, h5file=None, data_group=''):
        data_group = data_group + '/' + self.name
        self._logger.load(h5file, data_group)

    def save_log_file(self, save_dir=None):
        if save_dir is None:
            save_dir = './data/'
        os.makedirs(save_dir, exist_ok=True)
        file = h5py.File(save_dir + 'log.hdf5', 'w')
        self.save(file)
        file.close()

    def load_log_file(self, save_dir=None):
        if save_dir is None:
            save_dir = './data/'
        file = h5py.File(save_dir + 'log.hdf5', 'r')
        self.load(file)
        file.close()

    @property
    def state(self) -> Dict[str, np.ndarray]:
        return self._get_states()

    def _get_states(self) -> Dict[str, np.ndarray]:
        states = dict()
        for name, var in self.state_vars.items():
            states[name] = var.state
        return states

    @property
    def deriv(self) -> Dict[str, np.ndarray]:
        return self._get_deriv()

    def _get_deriv(self) -> Dict[str, np.ndarray]:
        derivs = dict()
        for name, var in self.state_vars.items():
            derivs[name] = var.deriv
        return derivs


class DynSystem(DynObject):
    def __init__(self, initial_states: Dict[str, ArrayType], deriv_fun=None, output_fun=None,
                 interval: Union[int, float] = -1):
        """ initial_states: dictionary of (state1, state2, ...) """
        super(DynSystem, self).__init__(initial_states=initial_states, interval=interval)
        self.name = "dyn_system"
        self.initial_states = initial_states

        if isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.evaluate
        else:
            self.deriv_fun = deriv_fun

        if isinstance(output_fun, BaseFunction):
            self.output_fun = output_fun.evaluate
        else:
            self.output_fun = output_fun

    # override
    def reset(self):
        super(DynSystem, self).reset()
        self.set_state(**self.initial_states)

    # may be implemented
    def _deriv(self, **kwargs) -> Dict[str, np.ndarray]:
        """ implement this method if needed
        args: dictionary of {name1: state1, name2: state2, ..., input_name1: input1, ...}
        return: dictionary of {name1: derivState1, name2: derivState2, ...)
        """
        if self.deriv_fun is None:
            raise NotImplementedError
        return self.deriv_fun(**kwargs)

    # override
    def forward(self, **kwargs):
        self._timer.forward()
        output = self._forward(**kwargs)

        if self._timer.is_event:
            self._last_output = output

        return self._last_output

    # implement
    def _forward(self, **kwargs):
        states = self._get_states()
        derivs = self._deriv(**states, **kwargs)

        for name, var in self.state_vars.items():
            var.set_deriv(derivs[name])

        if self._log_timer.is_event:
            self._logger.append(t=self.time, **states, **kwargs)
        return self._output()

    # override
    @property
    def output(self) -> Optional[np.ndarray]:
        return self._output()

    # may be overridden
    def _output(self) -> Optional[np.ndarray]:
        if self.output_fun is None:
            return None
        else:
            states = self._get_states()
            return self.output_fun(**states)


class TimeVaryingDynSystem(DynObject):
    def __init__(self, initial_states: Dict[str, ArrayType], deriv_fun=None, output_fun=None,
                 interval: Union[int, float] = -1):
        super(TimeVaryingDynSystem, self).__init__(initial_states=initial_states, interval=interval)
        self.name = "time_varying_dyn_system"
        self.initial_states = initial_states

        if isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.evaluate
        else:
            self.deriv_fun = deriv_fun

        if isinstance(output_fun, BaseFunction):
            self.output_fun = output_fun.evaluate
        else:
            self.output_fun = output_fun

    # override
    def reset(self):
        super(TimeVaryingDynSystem, self).reset()
        self.set_state(**self.initial_states)

    # to be implemented
    def _deriv(self, t, **kwargs) -> Dict[str, np.ndarray]:
        if self.deriv_fun is None:
            raise NotImplementedError
        return self.deriv_fun(t, **kwargs)

    # implement
    def _forward(self, **kwargs):
        states = self._get_states()
        derivs = self._deriv(t=self.time, **states, **kwargs)

        for name, var in self.state_vars.items():
            var.set_deriv(derivs[name])

        if self._log_timer.is_event:
            self._logger.append(t=self.time, **states, **kwargs)
        return self._output()

    # override
    @property
    def output(self) -> Optional[np.ndarray]:
        return self._output()

    # may be overridden
    def _output(self) -> Optional[np.ndarray]:
        if self.output_fun is None:
            return None
        else:
            states = self._get_states()
            return self.output_fun(t=self.time, **states)


class MultipleSystem(DynObject, ABC):
    def __init__(self, interval: Union[int, float] = -1):
        super(MultipleSystem, self).__init__(interval=interval)
        self.name = "model"
        self.sim_obj_list = []
        self.sim_obj_num = 0

    def attach_sim_objects(self, sim_obj_list: Union[SimObject, list, tuple]):
        if isinstance(sim_obj_list, SimObject):
            sim_obj_list = [sim_obj_list]

        for sim_obj in sim_obj_list:
            if not isinstance(sim_obj, SimObject):
                continue

            self.sim_obj_list.append(sim_obj)
            self.sim_obj_num += 1
            if isinstance(sim_obj, DynObject):
                for name, var in sim_obj.state_vars.items():
                    modified_name = 'sub' + str(self.sim_obj_num - 1) + '_' + name
                    self.state_vars[modified_name] = var

                self.num_state_var += sim_obj.num_state_var
                self.state_dim += sim_obj.state_dim

    # override
    def attach_sim_clock(self, sim_clock: SimClock):
        super(MultipleSystem, self).attach_sim_clock(sim_clock)
        for sim_obj in self.sim_obj_list:
            sim_obj.attach_sim_clock(sim_clock)

    # override
    def attach_log_timer(self, log_timer: Timer):
        super(MultipleSystem, self).attach_log_timer(log_timer)
        for sim_obj in self.sim_obj_list:
            sim_obj.attach_log_timer(log_timer)

    # override
    def reset(self):
        super(MultipleSystem, self).reset()
        for sim_obj in self.sim_obj_list:
            sim_obj.reset()

    # implement
    def check_stop_condition(self) -> Tuple[bool, list]:
        to_stop_list = []
        for sim_obj in self.sim_obj_list:
            to_stop_list.append(sim_obj.check_stop_condition())

        to_stop = np.array(to_stop_list, dtype=bool).any()
        if to_stop:
            self.flag = []
            for sim_obj in self.sim_obj_list:
                self.flag.append(sim_obj.flag)

        return to_stop, self.flag

    # implement
    def save(self, h5file=None, data_group=''):
        super().save(h5file, data_group)
        for sim_obj in self.sim_obj_list:
            if isinstance(sim_obj, DynObject):
                sim_obj.save(h5file, data_group + '/' + self.name)

    # implement
    def load(self, h5file=None, data_group=''):
        super().load(h5file, data_group)
        for sim_obj in self.sim_obj_list:
            if isinstance(sim_obj, DynObject):
                sim_obj.load(h5file, data_group + '/' + self.name)



import numpy as np
from pysimenv.core.system import MultipleSystem
from pysimenv.multicopter.model import MulticopterDynamic
from pysimenv.multicopter.control import QuaternionPosControl, QuaternionAttControl
from pysimenv.core.simulator import Simulator


class Model(MultipleSystem):
    def __init__(self):
        super(Model, self).__init__()

        m = 1.3
        J = np.diag([0.0119, 0.0119, 0.0219])

        pos = np.array([1., 2., 0.])
        vel = np.zeros(3)
        R_iv = np.identity(3)
        omega = np.zeros(3)
        self.quadrotor = MulticopterDynamic([pos, vel, R_iv, omega], m, J)

        K_att = QuaternionAttControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=1e-4*np.identity(3)
        )
        self.att_control = QuaternionAttControl(J, K_att)

        K_pos = QuaternionPosControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=np.identity(3)
        )
        self.pos_control = QuaternionPosControl(m, K_pos)

        self.attach_sim_objects([self.quadrotor, self.att_control, self.pos_control])

    def _forward(self, p_d: np.ndarray, v_d: np.ndarray = np.zeros(3)):
        p = self.quadrotor.pos
        v = self.quadrotor.vel
        q = self.quadrotor.quaternion
        omega = self.quadrotor.ang_vel

        f, q_d, omega_d = self.pos_control.forward(p, v, p_d, v_d)
        tau = self.att_control.forward(q, omega, q_d, omega_d)

        self.quadrotor.forward(u=np.array([f, tau[0], tau[1], tau[2]]))


def main():
    print("== Test for position control based on quaternion ==")
    p_d = np.array([0., 0., 1.])
    v_d = np.array([0., 0., 0.])
    model = Model()
    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, p_d=p_d, v_d=v_d)
    model.quadrotor.default_plot(show=True)


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pysimenv",
    version="0.0.5",
    author="Sangmin Lee",
    author_email="everlastingminii@gmail.com",
    description="A framework for performing numerical simulation of dynamic systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/minii93/python-sim-env",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'h5py',
        'pytictoc',
    ]
)



import numpy as np
from typing import Union, List, Tuple, Optional
from pysimenv.core.base import SimObject, StaticObject, ArrayType
from pysimenv.core.system import DynObject, MultipleSystem, DynSystem


class SignalGenerator(StaticObject):
    def __init__(self, shaping_fun, interval: Union[int, float] = -1):
        super(SignalGenerator, self).__init__(interval=interval)
        self.shaping_fun = shaping_fun

    # implementation
    def _forward(self):
        return self.shaping_fun(self._sim_clock.time)


class FeedbackControl(MultipleSystem):
    def __init__(self, system: DynObject, control: StaticObject):
        super(FeedbackControl, self).__init__()
        self.system = system
        self.control = control

        self.attach_sim_objects([system, control])

    # implement
    def _forward(self, r: Optional[np.ndarray] = None):
        y = self.system.output
        if r is None:
            u_fb = self.control.forward(y)
        else:
            u_fb = self.control.forward(y, r)
        self.system.forward(u=u_fb)


class Sequential(MultipleSystem):
    def __init__(self, obj_list: Union[List[SimObject], Tuple[SimObject]]):
        super(Sequential, self).__init__()
        assert len(obj_list) > 0, "(sequential) Invalid obj_list!"
        self.obj_list = list(obj_list)
        self.first_obj = obj_list[0]
        self.other_obj_list = obj_list[1:]

        self.attach_sim_objects(obj_list)

    # implement
    def _forward(self, **kwargs):
        self.first_obj.forward(**kwargs)
        out = self.first_obj.output
        for obj in self.other_obj_list:
            if out is None:
                obj.forward()
                out = obj.output
            elif isinstance(out, dict):
                obj.forward(**out)
                out = obj.output
            else:
                obj.forward(u=out)
                out = obj.output
        return out


class FlatEarthEnv(object):
    grav_accel = 9.805


class LinSys(DynSystem):
    def __init__(self, x_0: ArrayType, A: np.ndarray, B: Optional[np.ndarray] = None, C: Optional[np.ndarray] = None):
        super(LinSys, self).__init__(initial_states={'x': x_0})
        self.A = A
        self.B = B
        if C is None:
            C = np.eye(self.state_dim, dtype=np.float32)
        self.C = C

    # implement
    def _deriv(self, x: np.ndarray, u: Union[None, np.ndarray] = None):
        if self.B is None or u is None:
            x_dot = self.A.dot(x)
        else:
            x_dot = self.A.dot(x) + self.B.dot(u)

        return {'x': x_dot}

    # implement
    def _output(self) -> np.ndarray:
        return self.C.dot(self.state['x'])


class FirstOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, tau: float):
        A = np.array([[-1./tau]])
        B = np.array([[1./tau]])

        super(FirstOrderLinSys, self).__init__(x_0, A, B)
        self.tau = tau

    # implement
    def _output(self) -> Union[None, tuple, np.ndarray]:
        return self.state['x'][0]


class SecondOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, zeta: float, omega: float):
        A = np.array([
            [0., 1.],
            [-omega**2, -2*zeta*omega]
        ], dtype=np.float32)
        B = np.array([[0.], [omega**2]], dtype=np.float32)

        super(SecondOrderLinSys, self).__init__(x_0, A, B)
        self.zeta = zeta
        self.omega = omega

    # implement
    def _output(self) -> Union[None, tuple, np.ndarray]:
        return self.state['x'][0]


class Integrator(DynSystem):
    def __init__(self, initial_state: Union[ArrayType]):
        def deriv_fun(x: np.ndarray, u: Union[float, np.ndarray]):
            if isinstance(u, float):
                u = np.array([u])
            return {'x': u}

        super(Integrator, self).__init__({'x': initial_state}, deriv_fun)



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile2dof


print("== Test for PlanarMissile2dof ==")
initial_state = [-5000., 0., 200., np.deg2rad(30.)]
missile = PlanarMissile2dof(initial_state)

a_M = np.array([0., -5.])
dt = 0.01
time = 50.

simulator = Simulator(missile)
simulator.propagate(dt, time, True, a_M=a_M)

missile.report()
missile.plot()
missile.plot_path()
plt.show()



"""
Description for Package
"""
from pysimenv.missile.model import PlanarManVehicle2dof, PlanarNonManVehicle2dof, PlanarMissile2dof
from pysimenv.missile.engagement import Engagement2dim, PurePNG2dimEngagement, IACBPNGEngagement
from pysimenv.missile.guidance import PurePNG2dim, IACBPNG
from pysimenv.missile.util import RelKin2dim, CloseDistCond, miss_distance

__all__ = ['missile']